-------------------------------------------------------------------------------
--ResearchAssistantRemix.lua
--Author: tejn, based on ResearchAssistant by ingeniousclown
--v005
--[[
Research Assistant Remix puts a marker by items with traits.

When you own only one item with an unresearched trait, the marker is orange.
If you own multiple items with the same unresearched trait, one "preferred"
item is marked orange and the rest are dark green. If two or more items are
equally preferred, they are marked yellow. Preference is determined first by
lowest quality, then by lowest item level, then by whether the item is in
your bank.

If you already know an item's trait, it is marked gray with a different icon.
Intricate items are marked with a green diamond, and ornate items (including
jewelry!) are marked with a pile of gold. Other jewelry, gear without traits,
and all other items are not marked.

NOTE: Preferential marking works in the guild bank in comparison to items you
personally own, but NOT other guild bank contents. If multiple guild bank
items are better than what you have, they are all marked orange regardless of
how they compare to each other.
]]
-------------------------------------------------------------------------------

local BACKPACK = ZO_PlayerInventoryBackpack
local BANK = ZO_PlayerBankBackpack
local GUILD_BANK = ZO_GuildBankBackpack
local DECONSTRUCTION = ZO_SmithingTopLevelDeconstructionPanelInventoryBackpack

local CAN_RESEARCH_TEXTURE = [[/esoui/art/buttons/checkbox_indeterminate.dds]]
local IS_RESEARCHED_TEXTURE = [[/esoui/art/buttons/checkbox_checked_disabled.dds]]
local IS_INTRICATE_TEXTURE = [[/esoui/art/miscellaneous/bullet.dds]]
local IS_ORNATE_TEXTURE = [[/esoui/art/mappins/minimap_bank.dds]]
local CRAFTED_BY_PLAYER_TEXTURE = [[/esoui/art/buttons/decline_up.dds]] -- find this data!

local BLACKSMITH = CRAFTING_TYPE_BLACKSMITHING
local CLOTHIER = CRAFTING_TYPE_CLOTHIER
local WOODWORK = CRAFTING_TYPE_WOODWORKING

local controlsToWatch = {}
local ownedTraits = {}

local isScanning = false
local scanMore = 0


local function CreateItemPreferenceValue(bagId, slotIndex)
    local _,_,_,_,_,_,_,quality = GetItemInfo(bagId, slotIndex)
    if (not quality) then
        quality = 0
    end
    local level = GetItemLevel(bagId, slotIndex)
    if (not level) then
        level = 0
    end
    local where = 2
    if (bagId == BAG_BANK) then
        where = 0
    end
    if (bagId == BAG_BACKPACK) then
        where = 1
    end
    return quality*100000 + level*100 + where*10
    -- last digit reserved for tiebreakers in ScanBag
end

local function CreateIndicatorControl(parent)
	local control = WINDOW_MANAGER:CreateControl(parent:GetName() .. "Research", parent, CT_TEXTURE)
	control:SetDimensions(16, 16)
	control:SetAnchor(CENTER, parent, CENTER, 100)
	control:SetHidden(true)
	return control
end

local function CheckIsItemResearchableInSkill(bagId, slotIndex, equipType, craftingSkillType, traitIndex)
	--if it can't be extracted or refined here, then it can't be researched!
	if not CanItemBeSmithingExtractedOrRefined(bagId, slotIndex, craftingSkillType) then return nil end
	
	local numLines = GetNumSmithingResearchLines(craftingSkillType)

	for i=1, numLines do
		if (CanItemBeSmithingTraitResearched(bagId, slotIndex, craftingSkillType, i, traitIndex)) then
			return craftingSkillType * 1000000 + equipType * 10000 + i * 100 + traitIndex
		end
	end

	return nil  --nil will evaluate as false
end

local function CheckIsItemResearchable(bagId, slotIndex)
	local itemType = GetItemType(bagId, slotIndex)
	local _,_,_,_,_,equipType = GetItemInfo(bagId, slotIndex)
	if((itemType ~= ITEMTYPE_ARMOR and itemType ~= ITEMTYPE_WEAPON)
		or equipType == EQUIP_TYPE_NECK or equipType == EQUIP_TYPE_RING ) then
		return -1
	end

	local traitType = GetItemTrait(bagId, slotIndex)
	local traitIndex = traitType
	--this used to be "if(itemType == ITEMTYPE_ARMOR)", but shields are not armor even though they are armor
	if(traitIndex > 10) then
		traitIndex = traitIndex - 10;
	end

	if(not ((traitIndex >= 1 and traitIndex <=8) or (traitIndex >= 11 and traitIndex <= 18))) then
		return nil
	end

	return CheckIsItemResearchableInSkill(bagId, slotIndex, equipType, BLACKSMITH, traitIndex)
	 	or CheckIsItemResearchableInSkill(bagId, slotIndex, equipType, CLOTHIER, traitIndex)
        or CheckIsItemResearchableInSkill(bagId, slotIndex, equipType, WOODWORK, traitIndex)
end

local function AddResearchIndicatorToSlot(control)
	local bagId = control.dataEntry.data.bagId
	local slotIndex = control.dataEntry.data.slotIndex
	
	local indicatorControl = control:GetNamedChild("Research")
	if(not indicatorControl) then
		indicatorControl = CreateIndicatorControl(control)
		table.insert(controlsToWatch, indicatorControl)
	end

--[[
    local crafterName = ???
    if (crafterName and crafterName == GetUnitName("player")) then
        indicatorControl:SetTexture(CRAFTED_BY_PLAYER_TEXTURE)
        indicatorControl:SetColor(1, 0, 0, 1)
        indicatorControl:SetHidden(false)
        return
    end
]]    

	local traitKey = CheckIsItemResearchable(bagId, slotIndex)
    if (traitKey and traitKey < 0) then
        if (GetItemTrait(bagId, slotIndex) == ITEM_TRAIT_TYPE_JEWELRY_ORNATE) then
            indicatorControl:SetColor(1, 0.8, 0, 1)
            indicatorControl:SetTexture(IS_ORNATE_TEXTURE)
            indicatorControl:SetHidden(false)
        else
            indicatorControl:SetHidden(true)
        end
        return
    end

	if(traitKey) then
        indicatorControl:SetTexture(CAN_RESEARCH_TEXTURE)
        indicatorControl:SetHidden(false)
        local prefValue = CreateItemPreferenceValue(bagId, slotIndex)
		if(ownedTraits[traitKey] and ownedTraits[traitKey] < prefValue) then
            if (prefValue - ownedTraits[traitKey] < 10) then
                indicatorControl:SetColor(1, 0.9, 0.2, 1)
            else
                indicatorControl:SetColor(0, 0.5, 0, 1)
            end
		else
			indicatorControl:SetColor(1, 0.5, 0, 1)
		end
	else
        local traitType = GetItemTrait(bagId, slotIndex)
        if (not traitType or traitType < 1) then
            indicatorControl:SetHidden(true)
            return
        end
        indicatorControl:SetHidden(false)
        if (traitType == ITEM_TRAIT_TYPE_ARMOR_ORNATE
            or traitType == ITEM_TRAIT_TYPE_WEAPON_ORNATE) then
            indicatorControl:SetColor(1, 0.8, 0, 1)
            indicatorControl:SetTexture(IS_ORNATE_TEXTURE)
        elseif (traitType == ITEM_TRAIT_TYPE_ARMOR_INTRICATE
            or traitType == ITEM_TRAIT_TYPE_WEAPON_INTRICATE) then
            indicatorControl:SetColor(0.2, 1, 0.2, 1)
            indicatorControl:SetTexture(IS_INTRICATE_TEXTURE)
        else
            indicatorControl:SetColor(1, 1, 1, 1)
            indicatorControl:SetTexture(IS_RESEARCHED_TEXTURE)
        end
	end

end

local function AddResearchIndicators(self)
	for _,v in pairs(self.activeControls) do
		AddResearchIndicatorToSlot(v)
	end
end

local function CheckNow(self)
	if(#self.activeControls > 0 and not self.isGrid and not self:IsHidden()) then
        AddResearchIndicators(self)
    end
end

function AreAllHidden()
	return BANK:IsHidden() and BACKPACK:IsHidden() and GUILD_BANK:IsHidden() and DECONSTRUCTION:IsHidden()
end

local bufferTime = 50 --ms
local elapsedTime = 0
local stopUpdateWait = 500
local elapsedStopUpdateWait = 0
local shouldStopUpdate = false
function RA_OnUpdate()
	elapsedTime = elapsedTime + GetFrameDeltaTimeMilliseconds()
	if(isScanning or elapsedTime < bufferTime) then return end
	elapsedTime = 0

	if((AreAllHidden() and GetCraftingInteractionType() == 0) and not isStoreOrBankOpen) then
		shouldStopUpdate = true
		elapsedStopUpdateWait = elapsedStopUpdateWait + GetFrameDeltaTimeMilliseconds()
		if(elapsedStopUpdateWait >= stopUpdateWait) then
			RA_Controller:SetHandler("OnUpdate", nil)
			elapsedStopUpdateWait = 0
		end
		return
	end

	if(shouldStopUpdate) then
		shouldStopUpdate = false
		elapsedStopUpdateWait = 0
	end
    
	CheckNow(BANK)
	CheckNow(BACKPACK)
	CheckNow(GUILD_BANK)
	CheckNow(DECONSTRUCTION)
end

local function ScanBag( bagId )
	local _, numSlots = GetBagInfo(bagId)
	for i=1, numSlots do
		local traitKey = CheckIsItemResearchable(bagId, i)
		if(traitKey and traitKey >= 0) then
            local prefValue = CreateItemPreferenceValue(bagId, i)
			if(ownedTraits[traitKey]) then
                if (ownedTraits[traitKey] > prefValue) then
                    ownedTraits[traitKey] = prefValue
                else
                    if (ownedTraits[traitKey] == prefValue) then  -- duplicate result, deprecate both
                        ownedTraits[traitKey] = ownedTraits[traitKey] - 1
                    end
                end
			else
				ownedTraits[traitKey] = prefValue
			end
		end
	end
end

local function ResetOwnedTraitsTable()
	if(isScanning) then scanMore = scanMore + 1 end
	isScanning = true
	ownedTraits = {}
	ScanBag(BAG_BACKPACK)
	ScanBag(BAG_BANK)
	ScanBag(BAG_WORN)

	if(scanMore ~= 0) then
		scanMore = scanMore - 1
		ResetOwnedTraitsTable()
	else
		isScanning = false
	end
end

local function RA_InvUpdate( ... )
	ResetOwnedTraitsTable()
	if(not AreAllHidden()) then
		RA_Controller:SetHandler("OnUpdate", RA_OnUpdate)
	end
end

local function RA_ALPush( ... )
	RA_Controller:SetHandler("OnUpdate", RA_OnUpdate)
end

local function RA_StoreOrBankOpen( ... )
	RA_Controller:SetHandler("OnUpdate", RA_OnUpdate)
	isStoreOrBankOpen = true
end

local function RA_StoreOrBankClosed( ... )
	isStoreOrBankOpen = false
end

local function ResearchAssistant_Loaded(eventCode, addOnName)
	if(addOnName ~= "ResearchAssistantRemix") then
        return
    end

	EVENT_MANAGER:RegisterForEvent("RA_ALPUSH", EVENT_ACTION_LAYER_PUSHED, RA_ALPush)
	EVENT_MANAGER:RegisterForEvent("RA_INV_SLOT_UPDATE", EVENT_INVENTORY_SINGLE_SLOT_UPDATE, RA_InvUpdate)
    EVENT_MANAGER:RegisterForEvent("RA_STORE_OPEN", EVENT_OPEN_STORE, RA_StoreOrBankOpen)
	EVENT_MANAGER:RegisterForEvent("RA_BANK_OPEN", EVENT_OPEN_BANK, RA_StoreOrBankOpen)
	EVENT_MANAGER:RegisterForEvent("RA_STORE_CLOSE", EVENT_CLOSE_STORE, RA_StoreOrBankClosed)
	EVENT_MANAGER:RegisterForEvent("RA_BANK_CLOSE", EVENT_CLOSE_BANK, RA_StoreOrBankClosed)

    ResetOwnedTraitsTable()
end

local function ResearchAssistant_Initialized()
	EVENT_MANAGER:RegisterForEvent("ResearchAssistantLoaded", EVENT_ADD_ON_LOADED, ResearchAssistant_Loaded)
end

function RAT()
	d(ownedTraits)
end

ResearchAssistant_Initialized()